from capnp.includes.types cimport *
from capnp cimport helpers
from capnp.includes.capnp_cpp cimport DynamicValue, Schema, StringPtr
from capnp.lib.capnp cimport _DynamicStructReader, to_python_reader, _Schema

from capnp.helpers.non_circular cimport reraise_kj_exception

cdef extern from "capnp/helpers/capabilityHelper.cpp":
    pass

cdef extern from "graph_api.capnp.h":

    {%- for c in classes %}

    Schema get{{ c.cython_name }}_Schema"capnp::Schema::from<{{ c.cpp_name }}>"()
    cdef cppclass C_{{ c.cython_name }}_Reader "{{ c.cpp_name }}::Reader":
        {%- if c.fields == [] %}
        pass
        {%- endif %}
        {%- if c.union_fields != [] %}
        {{ c.cython_name }}_Which which() except +reraise_kj_exception
        {%- endif %}
        {%- for field in c.fields %}
        {{ field.cpp_return }} get{{ field.cpp_name }}() except +reraise_kj_exception
        {%- if field.union== True %}
        cbool is{{ field.cpp_name }}() except +reraise_kj_exception
        {%- endif %}
        {%- if field.gen_has == True %}
        cbool has{{ field.cpp_name }}() except +reraise_kj_exception
        {%- endif %}
        {%- endfor %}

    {%- if c.union_fields != [] %}
    cpdef enum class {{ c.cython_name }}_Which "{{ c.cpp_name }}::Which"(uint16_t):
        {%- for field in c.union_fields %}
        {{ field }},
        {%- endfor %}
    {%- endif %}
    {%- endfor %}

    {%- for e in enums %}
    cpdef enum class {{ e.cython_name }} "{{ e.cpp_name }}"(uint16_t):
        {%- for enumerant in e.enumerants %}
        {{ enumerant }},
        {%- endfor %}
    {%- endfor %}

    cdef cppclass C_DynamicStruct_Reader" ::capnp::DynamicStruct::Reader":
        {%- for c in classes %}
        C_{{ c.cython_name }}_Reader as{{ c.cython_name }}"as<{{ c.cpp_name }}>"()
        {%- endfor %}

cdef extern from "capnp/list.h":
    cdef cppclass C_Uint16_List " ::capnp::List<uint16_t, ::capnp::Kind::PRIMITIVE>::Reader":
        uint16_t operator[](uint) except +reraise_kj_exception
        uint size()
    cdef cppclass C_Uint32_List " ::capnp::List<uint32_t, ::capnp::Kind::PRIMITIVE>::Reader":
        uint32_t operator[](uint) except +reraise_kj_exception
        uint size()
    cdef cppclass C_Uint64_List " ::capnp::List<uint64_t, ::capnp::Kind::PRIMITIVE>::Reader":
        uint64_t operator[](uint) except +reraise_kj_exception
        uint size()
    cdef cppclass C_Int64_List " ::capnp::List<int64_t, ::capnp::Kind::PRIMITIVE>::Reader":
        int64_t operator[](uint) except +reraise_kj_exception
        uint size()
    cdef cppclass C_String_List " ::capnp::List<capnp::Text, ::capnp::Kind::BLOB>::Reader":
        StringPtr operator[](uint) except +reraise_kj_exception
        uint size()

    {%- for c in classes_lists %}
    cdef cppclass C_{{ c.cython_name }}_List " ::capnp::List<{{ c.cpp_name }}, ::capnp::Kind::STRUCT>::Reader":
        C_{{ c.cython_name }} operator[](uint) except +reraise_kj_exception
        uint size()
    {%- endfor %}

{%- for c in classes %}

cdef class {{ c.cython_name }}_Reader:
    cdef C_{{ c.cython_name }}_Reader source
    cdef object root
    @staticmethod
    cdef init(C_{{ c.cython_name }}_Reader source, object root)
{%- endfor %}

cdef class Uint16_List:
    cdef C_Uint16_List source
    cdef object root
    @staticmethod
    cdef init(C_Uint16_List source, object root)

cdef class Uint32_List:
    cdef C_Uint32_List source
    cdef object root
    @staticmethod
    cdef init(C_Uint32_List source, object root)

cdef class Uint64_List:
    cdef C_Uint64_List source
    cdef object root
    @staticmethod
    cdef init(C_Uint64_List source, object root)

cdef class Int64_List:
    cdef C_Int64_List source
    cdef object root
    @staticmethod
    cdef init(C_Int64_List source, object root)

cdef class String_List:
    cdef C_String_List source
    cdef object root
    @staticmethod
    cdef init(C_String_List source, object root)

{%- for c in classes_lists %}
cdef class {{ c.cython_name }}_List:
    cdef C_{{ c.cython_name }}_List source
    cdef object root
    @staticmethod
    cdef init(C_{{ c.cython_name }}_List source, object root)
{%- endfor %}



import argparse
from multiprocessing import Pool
from pathlib import Path
from functools import partial
from collections import Counter
import math
from pytact.data_reader import GlobalContextSets, lowlevel_data_reader, lowlevel_to_highlevel, Definition, Node, Original, Substituted, Discharged, node_dependencies, definition_dependencies
import pytact.graph_api_capnp_cython as apic
import capnp
import pytact.graph_api_capnp as graph_api_capnp

def open_dataset(dataset_path: Path):
    global ctx
    ctx = lowlevel_data_reader(dataset_path)
    global data
    data = ctx.__enter__()
    global node_counts
    node_counts = [len(data[i].graph.nodes) for i in range(0, len(data.graph_files))]
    global hdata
    hdata = lowlevel_to_highlevel(data)
    global global_contexts
    global_contexts = GlobalContextSets.new_context().__enter__()

# def node_dependencies(n: Node, deps: set[Definition] | None = None) -> set[Definition]:
#     if deps is None:
#         deps = set()
#     seen: set[Node] = set()
#     stack = [n]
#     while stack:
#         node = stack.pop()
#         if node in seen:
#             continue
#         seen.add(node)
#         if d := node.definition:
#             deps.add(d)
#             continue
#         for _, child in node.children:
#             stack.append(child)
#     return deps

# def definition_dependencies(d: Definition):
#     deps = set()
#     for _, c in d.node.children:
#         node_dependencies(c, deps)
#     return deps

def process1(args, fname: Path):
    file_errors = []
    file_definitions = 0
    file_original_definitions = 0
    file_base_tactics_text = set()
    file_base_tactics_intermtext = set()
    file_tactics = Counter()
    file_original_tactics = Counter()
    file_tactic_arguments = {}
    file_tactic_ident_to_base = {}
    proof_steps = 0
    original_proof_steps = 0
    original_proof_steps_faithful = 0
    outcomes = 0
    original_outcomes = 0
    proofs = 0
    original_proofs = 0
    original_proofs_faithful = 0
    unresolvable = 0
    file_disappearing_proof_states = 0
    graphid = data.graphid_by_filename[fname]
    freader = data[graphid]
    graph = freader.graph
    nodes_count = node_counts[graphid]
    edges_count = len(graph.edges)
    dependency_count = len(freader.dependencies)
    print(f"Checking file {fname}: {nodes_count} nodes and {edges_count} edges")

    # Check correctness of Graph struct
    for n in graph.nodes:
        if not (n.children_index + n.children_count <= edges_count):
            file_errors.append(f"{fname}: Children of node {n} are not valid indices of Graph.edges")
        if n.label.is_definition:
            file_definitions += 1

    for t in graph.edges:
        target = t.target
        dep_index = target.dep_index
        if dep_index >= dependency_count:
            file_errors.append(f"{fname}: target.depIndex {t.target.dep_index} "
                  f"but len(g.dependencies) is {dependency_count} "
                  f"and g.dependencies = {freader.dependencies}")
        if target.node_index >= node_counts[data.local_to_global(graphid, dep_index)]:
            file_errors.append(f"{fname}: Reference for target {t} is out of bounds")

    # Check correctness of Dataset struct
    if file_definitions != len(freader.definitions):
        file_errors.append(f"{fname}: Counted {file_definitions} definitions in the file, "
                        f"but Dataset.definitions has size {len(freader.definitions)}")

    hreader = hdata[fname]
    with global_contexts.sub_context(lambda d: d.is_file_representative) as sub_global_contexts:
        for d in hreader.definitions():
            if not d.node.definition:
                file_errors.append(f"{fname}: Node {d.node} should be a definition but is not")

            # Check correctness of Definition struct
            match d.status:
                case Original():
                    file_original_definitions += 1
                case Discharged(original):
                    if not original.node.definition:
                        file_errors.append(f"{fname}: Discharged node of definition {d.name} "
                                        f"is not a definition")
                case Substituted(original):
                    if not original.node.definition:
                        file_errors.append(f"{fname}: Substituted node of definition {d.name} "
                                        f"is not a definition")

            def check_in_global_context(s):
                global_context = sub_global_contexts.global_context_set(d)
                for dd in s:
                    if dd not in global_context:
                        file_errors.append(f"{fname}: Definition {d.name} has dependency {dd.name} "
                                        f"which is not part of its global context")

            if not args.quick:
                check_in_global_context(definition_dependencies(d))

            crepr = d.cluster_representative
            if d not in d.cluster:
                file_errors.append(f"{fname}: Cluster represented by {crepr.name}"
                                f"does not contain {d.name}")
            for cd in d.cluster:
                if crepr != cd.cluster_representative:
                    file_errors.append(f"{fname}: Cluster represented by {crepr.name} contains unrelated "
                                    f"definition {cd.name}")
            if prev := d.previous:
                if not prev.node.definition:
                    file_errors.append(f"{fname}: Previous node of definition {d.name} is not a definition")
            for ep in d.external_previous:
                if not ep.node.definition:
                    file_errors.append(f"{fname}: External dependency of definition {d.name} is "
                                    f"not a definition")
            if ps := d.proof:
                proofs += 1
                faithful = True
                before_states = set()
                for p in ps:
                    for outcome in p.outcomes:
                        before_states.add(outcome.before.id)
                for p in ps:
                    proof_steps += 1
                    if t := p.tactic:
                        ident = t.ident
                        file_base_tactics_text.add(t.base_text)
                        file_base_tactics_intermtext.add(t.interm_text)
                    else:
                        ident = 0
                    file_tactics[ident] += 1
                    if isinstance(d.status, Original):
                        file_original_tactics[ident] += 1
                        original_proof_steps += 1
                        if (t := p.tactic) and t.text == t.interm_text:
                                original_proof_steps_faithful += 1
                        else:
                            faithful = False
                    for outcome in p.outcomes:
                        for after in outcome.after:
                            if after.id not in before_states:
                                file_disappearing_proof_states += 1

                        if not args.quick:
                            check_in_global_context(node_dependencies(outcome.term))
                        for state in [outcome.before] + list(outcome.after):
                            root_label = state.root.label
                            if not root_label.is_proof_state:
                                file_errors.append(f"{fname}: root is proof state {state} is {root_label}")

                            if not args.quick:
                                check_in_global_context(node_dependencies(state.root))
                            if len(state.context) != len(state.context_names):
                                file_errors.append(f"{fname}: Length of context is different from length of"
                                                f"context_names in proof state {state}")
                            if len(state.context) != len(state.context_text):
                                file_errors.append(f"{fname}: Length of context is different from length of"
                                                f"context_text in proof state {state}")
                            root_children = [child for _, child in state.root.children]
                            for c in state.context:
                                if c not in root_children:
                                    file_errors.append(
                                        f"{fname}: hyp {c} of state {state} in def {d.name} is not "
                                        f"reachable from the root")
                                if not (c.label.is_context_def or c.label.is_context_assum):
                                    file_errors.append(
                                        f"{fname}: ctx {state.context} of a state {state} "
                                        f"has the wrong node classification {c.label}")

                        file_tactic_arguments.setdefault(ident, len(outcome.tactic_arguments))
                        if t := p.tactic:
                            file_tactic_ident_to_base.setdefault(ident, t.base_text)
                        if file_tactic_arguments[ident] != len(outcome.tactic_arguments):
                            file_errors.append(f"{fname}: Tactic with two different argument lengths detected. : Original: {file_tactic_arguments[ident]} : {file_tactic_ident_to_base[ident]}. New {len(outcome.tactic_arguments)} : {t.text} : {t.base_text}")
                        outcomes += 1
                        if isinstance(d.status, Original):
                            original_outcomes += 1
                        for a in outcome.tactic_arguments:
                            if not a:
                                unresolvable += 1
                            else:
                                if not args.quick:
                                    check_in_global_context(node_dependencies(a))
                if isinstance(d.status, Original):
                    original_proofs += 1
                    if faithful:
                        original_proofs_faithful += 1
    if freader.representative != len(graph.nodes):
        if not graph.nodes[freader.representative].label.is_definition:
            file_errors.append(f"{fname}: Representative {freader.representative} is not a definition")
    for dep in freader.dependencies:
        if Path(dep) not in data.graphid_by_filename:
            file_errors.append(f"{fname}: Dependency {dep} could not be found")


    return (fname, nodes_count, edges_count,
            file_definitions, file_original_definitions, file_base_tactics_text,
            file_base_tactics_intermtext, file_tactics, file_original_tactics, file_tactic_arguments,
            proof_steps, original_proof_steps, original_proof_steps_faithful,
            outcomes, original_outcomes, proofs, original_proofs,
            original_proofs_faithful, unresolvable, file_errors, file_disappearing_proof_states)

def entropy(d):
    n = sum(d.values())
    return -sum([(c / n) * math.log(c / n, 2) for c in d.values()])

def main2():
    parser = argparse.ArgumentParser(
        description = 'Run sanity checks on a dataset and print some statistics.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('dataset',
                        type=str,
                        help=('The location of the dataset to check. ' +
                              'Either a dataset directory, or a SquashFS image, ' +
                              'which will be automatically mounted.'))
    parser.add_argument('--jobs',
                        type=int,
                        default=None,
                        help='number of parallel multiprocessing jobs to run')
    parser.add_argument('--quick',
                        action='store_true',
                        help='skip the more expensive checks')
    parser.add_argument('--verbose',
                       type = int,
                       default = 0,
                       help='level of being verbose 0,1..')

    args = parser.parse_args()

    import sys
    sys.setrecursionlimit(10000)

    dataset_path = Path(args.dataset).resolve()

    errors = []
    tactics = Counter()
    original_tactics = Counter()
    tactic_arguments = {}
    base_tactics_text = set()
    base_tactics_intermtext = set()
    definitions = 0
    original_definitions = 0
    nodes_total = 0
    edges_total = 0
    proof_steps_total = 0
    original_proof_steps_total = 0
    original_proof_steps_faithful_total = 0
    outcomes_total = 0
    original_outcomes_total = 0
    proofs_total = 0
    original_proofs_total = 0
    original_proofs_faithful_total = 0
    unresolvable_total = 0
    disappearing_proof_states = 0

    with lowlevel_data_reader(dataset_path) as data:
        file_list = data.graph_files

    process1_partial = partial(process1, args)
    with Pool(args.jobs, open_dataset, [dataset_path]) as pool:
        results = pool.map(process1_partial, file_list, chunksize=1)
    # file_list = file_list[280:360]
    # print(file_list)
    # open_dataset(dataset_path)
    # results = [process1(args, f) for f in file_list]

    for res in results:
        (fname, nodes_count, edges_count,
         file_definitions, file_original_definitions, file_base_tactics_text,
         file_base_tactics_intermtext, file_tactics, file_original_tactics, file_tactic_arguments,
         proof_steps, original_proof_steps, original_proof_steps_faithful,
         outcomes, original_outcomes, proofs, original_proofs,
         original_proofs_faithful, unresolvable, file_errors, file_disappearing_proof_states) = res
        nodes_total += nodes_count
        edges_total += edges_count
        proof_steps_total += proof_steps
        original_proof_steps_total += original_proof_steps
        original_proof_steps_faithful_total += original_proof_steps_faithful
        outcomes_total += outcomes
        original_outcomes_total += original_outcomes
        proofs_total += proofs
        original_proofs_total += original_proofs
        original_proofs_faithful_total += original_proofs_faithful
        definitions +=  file_definitions
        original_definitions += file_original_definitions
        base_tactics_text.update(file_base_tactics_text)
        base_tactics_intermtext.update(file_base_tactics_intermtext)
        tactics += file_tactics
        original_tactics += file_original_tactics
        disappearing_proof_states += file_disappearing_proof_states
        errors += file_errors
        for tac, length in file_tactic_arguments.items():
            tactic_arguments.setdefault(tac, length)
            if tactic_arguments[tac] != length:
                errors.append(f'{fname}: Tactic with two different argument lengths detected: {tac}')

        unresolvable_total += unresolvable

    for error in errors:
        print(error)

    print(f"Errors encountered {len(errors)}")
    print(f"Nodes total {nodes_total}")
    print(f"Edges total {edges_total}")
    print(f"Tactics total {len(tactics)}")
    print(f"Original tactics total {len(original_tactics)}")
    print(f"Tactics entropy (bits) {entropy(tactics)}")
    print(f"Original tactics entropy (bits) {entropy(original_tactics)}")
    print(f"Tactics base text total {len(base_tactics_text)}")
    print(f"Tactics base intermtext total {len(base_tactics_intermtext)}")
    print(f"Definitions total {definitions}")
    print(f"Original definitions total {original_definitions}")
    print(f"Proof steps total {proof_steps_total}")
    print(f"Original proof steps total {original_proof_steps_total}")
    print(f"Faithfully represented original proof steps total {original_proof_steps_faithful_total}")
    print(f"Outcomes total {outcomes_total}")
    print(f"Original outcomes total {original_outcomes_total}")
    print(f"Proofs total {proofs_total}")
    print(f"Original proofs total {original_proofs_total}")
    print(f"Faithful original proofs total {original_proofs_faithful_total}")
    print(f"Unresolvable tactic arguments {unresolvable_total}")
    print(f"Disappearing proof states (spooky action at a distance) {disappearing_proof_states}")

    if (args.verbose >= 1):
        print("Tactics base text:")
        for t in base_tactics_text:
            print(t)
        print("Tactics intermtext text:")
        for t in base_tactics_intermtext:
            print(t)

    if len(errors) > 0:
        raise Exception("Errors occurred")

def main():
    # import pstats, cProfile
    # cProfile.runctx("main2()", globals(), locals(), "Profile.prof")
    # s = pstats.Stats("Profile.prof")
    # s.strip_dirs().sort_stats("time").print_stats()
    main2()

if __name__ == '__main__':
    main()



from capnp.includes.types cimport *
from capnp cimport helpers
from capnp.includes.capnp_cpp cimport DynamicValue, Schema, StringPtr
from capnp.lib.capnp cimport _DynamicStructReader, to_python_reader, _Schema

from capnp.helpers.non_circular cimport reraise_kj_exception

cdef extern from "capnp/helpers/capabilityHelper.cpp":
    pass

cdef extern from "graph_api.capnp.h":

    Schema getGraph_Node_Schema"capnp::Schema::from<Graph::Node>"()
    cdef cppclass C_Graph_Node_Reader "Graph::Node::Reader":
        C_Graph_Node_Label_Reader getLabel() except +reraise_kj_exception
        uint32_t getChildrenIndex() except +reraise_kj_exception
        uint16_t getChildrenCount() except +reraise_kj_exception
        int64_t getIdentity() except +reraise_kj_exception

    Schema getGraph_EdgeTarget_Schema"capnp::Schema::from<Graph::EdgeTarget>"()
    cdef cppclass C_Graph_EdgeTarget_Reader "Graph::EdgeTarget::Reader":
        EdgeClassification getLabel() except +reraise_kj_exception
        C_Graph_EdgeTarget_Target_Reader getTarget() except +reraise_kj_exception

    Schema getGraph_Schema"capnp::Schema::from<Graph>"()
    cdef cppclass C_Graph_Reader "Graph::Reader":
        C_Graph_Node_Reader_List getNodes() except +reraise_kj_exception
        cbool hasNodes() except +reraise_kj_exception
        C_Graph_EdgeTarget_Reader_List getEdges() except +reraise_kj_exception
        cbool hasEdges() except +reraise_kj_exception

    Schema getIntP_Schema"capnp::Schema::from<IntP>"()
    cdef cppclass C_IntP_Reader "IntP::Reader":
        uint64_t getValue() except +reraise_kj_exception

    Schema getDefinition_Schema"capnp::Schema::from<Definition>"()
    cdef cppclass C_Definition_Reader "Definition::Reader":
        Definition_Which which() except +reraise_kj_exception
        StringPtr getName() except +reraise_kj_exception
        cbool hasName() except +reraise_kj_exception
        uint32_t getPrevious() except +reraise_kj_exception
        C_Uint16_List getExternalPrevious() except +reraise_kj_exception
        cbool hasExternalPrevious() except +reraise_kj_exception
        C_Definition_Status_Reader getStatus() except +reraise_kj_exception
        uint32_t getInductive() except +reraise_kj_exception
        cbool isInductive() except +reraise_kj_exception
        uint32_t getConstructor() except +reraise_kj_exception
        cbool isConstructor() except +reraise_kj_exception
        uint32_t getProjection() except +reraise_kj_exception
        cbool isProjection() except +reraise_kj_exception
        void getManualConstant() except +reraise_kj_exception
        cbool isManualConstant() except +reraise_kj_exception
        C_ProofStep_Reader_List getTacticalConstant() except +reraise_kj_exception
        cbool isTacticalConstant() except +reraise_kj_exception
        cbool hasTacticalConstant() except +reraise_kj_exception
        void getManualSectionConstant() except +reraise_kj_exception
        cbool isManualSectionConstant() except +reraise_kj_exception
        C_ProofStep_Reader_List getTacticalSectionConstant() except +reraise_kj_exception
        cbool isTacticalSectionConstant() except +reraise_kj_exception
        cbool hasTacticalSectionConstant() except +reraise_kj_exception
        StringPtr getTypeText() except +reraise_kj_exception
        cbool hasTypeText() except +reraise_kj_exception
        StringPtr getTermText() except +reraise_kj_exception
        cbool hasTermText() except +reraise_kj_exception
    cpdef enum class Definition_Which "Definition::Which"(uint16_t):
        INDUCTIVE,
        CONSTRUCTOR,
        PROJECTION,
        MANUAL_CONSTANT,
        TACTICAL_CONSTANT,
        MANUAL_SECTION_CONSTANT,
        TACTICAL_SECTION_CONSTANT,

    Schema getFloatP_Schema"capnp::Schema::from<FloatP>"()
    cdef cppclass C_FloatP_Reader "FloatP::Reader":
        double getValue() except +reraise_kj_exception

    Schema getGraph_Node_Label_Schema"capnp::Schema::from<Graph::Node::Label>"()
    cdef cppclass C_Graph_Node_Label_Reader "Graph::Node::Label::Reader":
        Graph_Node_Label_Which which() except +reraise_kj_exception
        C_IntP_Reader getProofState() except +reraise_kj_exception
        cbool isProofState() except +reraise_kj_exception
        cbool hasProofState() except +reraise_kj_exception
        void getContextDef() except +reraise_kj_exception
        cbool isContextDef() except +reraise_kj_exception
        void getContextAssum() except +reraise_kj_exception
        cbool isContextAssum() except +reraise_kj_exception
        C_Definition_Reader getDefinition() except +reraise_kj_exception
        cbool isDefinition() except +reraise_kj_exception
        cbool hasDefinition() except +reraise_kj_exception
        void getConstEmpty() except +reraise_kj_exception
        cbool isConstEmpty() except +reraise_kj_exception
        void getSortSProp() except +reraise_kj_exception
        cbool isSortSProp() except +reraise_kj_exception
        void getSortProp() except +reraise_kj_exception
        cbool isSortProp() except +reraise_kj_exception
        void getSortSet() except +reraise_kj_exception
        cbool isSortSet() except +reraise_kj_exception
        void getSortType() except +reraise_kj_exception
        cbool isSortType() except +reraise_kj_exception
        void getRel() except +reraise_kj_exception
        cbool isRel() except +reraise_kj_exception
        void getEvar() except +reraise_kj_exception
        cbool isEvar() except +reraise_kj_exception
        void getEvarSubst() except +reraise_kj_exception
        cbool isEvarSubst() except +reraise_kj_exception
        void getCast() except +reraise_kj_exception
        cbool isCast() except +reraise_kj_exception
        void getProd() except +reraise_kj_exception
        cbool isProd() except +reraise_kj_exception
        void getLambda() except +reraise_kj_exception
        cbool isLambda() except +reraise_kj_exception
        void getLetIn() except +reraise_kj_exception
        cbool isLetIn() except +reraise_kj_exception
        void getApp() except +reraise_kj_exception
        cbool isApp() except +reraise_kj_exception
        void getCase() except +reraise_kj_exception
        cbool isCase() except +reraise_kj_exception
        void getCaseBranch() except +reraise_kj_exception
        cbool isCaseBranch() except +reraise_kj_exception
        void getFix() except +reraise_kj_exception
        cbool isFix() except +reraise_kj_exception
        void getFixFun() except +reraise_kj_exception
        cbool isFixFun() except +reraise_kj_exception
        void getCoFix() except +reraise_kj_exception
        cbool isCoFix() except +reraise_kj_exception
        void getCoFixFun() except +reraise_kj_exception
        cbool isCoFixFun() except +reraise_kj_exception
        C_IntP_Reader getInt() except +reraise_kj_exception
        cbool isInt() except +reraise_kj_exception
        cbool hasInt() except +reraise_kj_exception
        C_FloatP_Reader getFloat() except +reraise_kj_exception
        cbool isFloat() except +reraise_kj_exception
        cbool hasFloat() except +reraise_kj_exception
        StringPtr getPrimitive() except +reraise_kj_exception
        cbool isPrimitive() except +reraise_kj_exception
        cbool hasPrimitive() except +reraise_kj_exception
    cpdef enum class Graph_Node_Label_Which "Graph::Node::Label::Which"(uint16_t):
        PROOF_STATE,
        CONTEXT_DEF,
        CONTEXT_ASSUM,
        DEFINITION,
        CONST_EMPTY,
        SORT_S_PROP,
        SORT_PROP,
        SORT_SET,
        SORT_TYPE,
        REL,
        EVAR,
        EVAR_SUBST,
        CAST,
        PROD,
        LAMBDA,
        LET_IN,
        APP,
        CASE,
        CASE_BRANCH,
        FIX,
        FIX_FUN,
        CO_FIX,
        CO_FIX_FUN,
        INT,
        FLOAT,
        PRIMITIVE,

    Schema getDefinition_Status_Substituted_Schema"capnp::Schema::from<Definition::Status::Substituted>"()
    cdef cppclass C_Definition_Status_Substituted_Reader "Definition::Status::Substituted::Reader":
        uint16_t getDepIndex() except +reraise_kj_exception
        uint32_t getNodeIndex() except +reraise_kj_exception

    Schema getDefinition_Status_Schema"capnp::Schema::from<Definition::Status>"()
    cdef cppclass C_Definition_Status_Reader "Definition::Status::Reader":
        Definition_Status_Which which() except +reraise_kj_exception
        void getOriginal() except +reraise_kj_exception
        cbool isOriginal() except +reraise_kj_exception
        uint32_t getDischarged() except +reraise_kj_exception
        cbool isDischarged() except +reraise_kj_exception
        C_Definition_Status_Substituted_Reader getSubstituted() except +reraise_kj_exception
        cbool isSubstituted() except +reraise_kj_exception
    cpdef enum class Definition_Status_Which "Definition::Status::Which"(uint16_t):
        ORIGINAL,
        DISCHARGED,
        SUBSTITUTED,

    Schema getProofStep_Schema"capnp::Schema::from<ProofStep>"()
    cdef cppclass C_ProofStep_Reader "ProofStep::Reader":
        C_ProofStep_Tactic_Reader getTactic() except +reraise_kj_exception
        C_Outcome_Reader_List getOutcomes() except +reraise_kj_exception
        cbool hasOutcomes() except +reraise_kj_exception

    Schema getTactic_Schema"capnp::Schema::from<Tactic>"()
    cdef cppclass C_Tactic_Reader "Tactic::Reader":
        int64_t getIdent() except +reraise_kj_exception
        StringPtr getText() except +reraise_kj_exception
        cbool hasText() except +reraise_kj_exception
        StringPtr getBaseText() except +reraise_kj_exception
        cbool hasBaseText() except +reraise_kj_exception
        StringPtr getIntermText() except +reraise_kj_exception
        cbool hasIntermText() except +reraise_kj_exception
        cbool getExact() except +reraise_kj_exception
        StringPtr getTextNonAnonymous() except +reraise_kj_exception
        cbool hasTextNonAnonymous() except +reraise_kj_exception

    Schema getProofStep_Tactic_Schema"capnp::Schema::from<ProofStep::Tactic>"()
    cdef cppclass C_ProofStep_Tactic_Reader "ProofStep::Tactic::Reader":
        ProofStep_Tactic_Which which() except +reraise_kj_exception
        void getUnknown() except +reraise_kj_exception
        cbool isUnknown() except +reraise_kj_exception
        C_Tactic_Reader getKnown() except +reraise_kj_exception
        cbool isKnown() except +reraise_kj_exception
        cbool hasKnown() except +reraise_kj_exception
    cpdef enum class ProofStep_Tactic_Which "ProofStep::Tactic::Which"(uint16_t):
        UNKNOWN,
        KNOWN,

    Schema getOutcome_Schema"capnp::Schema::from<Outcome>"()
    cdef cppclass C_Outcome_Reader "Outcome::Reader":
        C_ProofState_Reader getBefore() except +reraise_kj_exception
        cbool hasBefore() except +reraise_kj_exception
        C_ProofState_Reader_List getAfter() except +reraise_kj_exception
        cbool hasAfter() except +reraise_kj_exception
        C_Outcome_Term_Reader getTerm() except +reraise_kj_exception
        StringPtr getTermText() except +reraise_kj_exception
        cbool hasTermText() except +reraise_kj_exception
        C_Argument_Reader_List getTacticArguments() except +reraise_kj_exception
        cbool hasTacticArguments() except +reraise_kj_exception

    Schema getOutcome_Term_Schema"capnp::Schema::from<Outcome::Term>"()
    cdef cppclass C_Outcome_Term_Reader "Outcome::Term::Reader":
        uint16_t getDepIndex() except +reraise_kj_exception
        uint32_t getNodeIndex() except +reraise_kj_exception

    Schema getProofState_Schema"capnp::Schema::from<ProofState>"()
    cdef cppclass C_ProofState_Reader "ProofState::Reader":
        C_ProofState_Root_Reader getRoot() except +reraise_kj_exception
        C_Node_Reader_List getContext() except +reraise_kj_exception
        cbool hasContext() except +reraise_kj_exception
        C_String_List getContextNames() except +reraise_kj_exception
        cbool hasContextNames() except +reraise_kj_exception
        C_String_List getContextText() except +reraise_kj_exception
        cbool hasContextText() except +reraise_kj_exception
        StringPtr getConclusionText() except +reraise_kj_exception
        cbool hasConclusionText() except +reraise_kj_exception
        StringPtr getText() except +reraise_kj_exception
        cbool hasText() except +reraise_kj_exception
        uint32_t getId() except +reraise_kj_exception

    Schema getArgument_Schema"capnp::Schema::from<Argument>"()
    cdef cppclass C_Argument_Reader "Argument::Reader":
        Argument_Which which() except +reraise_kj_exception
        void getUnresolvable() except +reraise_kj_exception
        cbool isUnresolvable() except +reraise_kj_exception
        C_Argument_Term_Reader getTerm() except +reraise_kj_exception
        cbool isTerm() except +reraise_kj_exception
    cpdef enum class Argument_Which "Argument::Which"(uint16_t):
        UNRESOLVABLE,
        TERM,

    Schema getProofState_Root_Schema"capnp::Schema::from<ProofState::Root>"()
    cdef cppclass C_ProofState_Root_Reader "ProofState::Root::Reader":
        uint16_t getDepIndex() except +reraise_kj_exception
        uint32_t getNodeIndex() except +reraise_kj_exception

    Schema getNode_Schema"capnp::Schema::from<Node>"()
    cdef cppclass C_Node_Reader "Node::Reader":
        uint16_t getDepIndex() except +reraise_kj_exception
        uint32_t getNodeIndex() except +reraise_kj_exception

    Schema getArgument_Term_Schema"capnp::Schema::from<Argument::Term>"()
    cdef cppclass C_Argument_Term_Reader "Argument::Term::Reader":
        uint16_t getDepIndex() except +reraise_kj_exception
        uint32_t getNodeIndex() except +reraise_kj_exception

    Schema getGraph_EdgeTarget_Target_Schema"capnp::Schema::from<Graph::EdgeTarget::Target>"()
    cdef cppclass C_Graph_EdgeTarget_Target_Reader "Graph::EdgeTarget::Target::Reader":
        uint16_t getDepIndex() except +reraise_kj_exception
        uint32_t getNodeIndex() except +reraise_kj_exception

    Schema getAbstractTactic_Schema"capnp::Schema::from<AbstractTactic>"()
    cdef cppclass C_AbstractTactic_Reader "AbstractTactic::Reader":
        int64_t getIdent() except +reraise_kj_exception
        uint8_t getParameters() except +reraise_kj_exception

    Schema getDataVersion_Schema"capnp::Schema::from<DataVersion>"()
    cdef cppclass C_DataVersion_Reader "DataVersion::Reader":
        int64_t getMajor() except +reraise_kj_exception

    Schema getDataset_Schema"capnp::Schema::from<Dataset>"()
    cdef cppclass C_Dataset_Reader "Dataset::Reader":
        C_String_List getDependencies() except +reraise_kj_exception
        cbool hasDependencies() except +reraise_kj_exception
        C_Graph_Reader getGraph() except +reraise_kj_exception
        cbool hasGraph() except +reraise_kj_exception
        uint32_t getRepresentative() except +reraise_kj_exception
        C_Uint32_List getDefinitions() except +reraise_kj_exception
        cbool hasDefinitions() except +reraise_kj_exception
        StringPtr getModuleName() except +reraise_kj_exception
        cbool hasModuleName() except +reraise_kj_exception
        C_DataVersion_Reader getDataVersion() except +reraise_kj_exception
        cbool hasDataVersion() except +reraise_kj_exception

    Schema getException_Schema"capnp::Schema::from<Exception>"()
    cdef cppclass C_Exception_Reader "Exception::Reader":
        Exception_Which which() except +reraise_kj_exception
        void getNoSuchTactic() except +reraise_kj_exception
        cbool isNoSuchTactic() except +reraise_kj_exception
        void getMismatchedArguments() except +reraise_kj_exception
        cbool isMismatchedArguments() except +reraise_kj_exception
        void getParseError() except +reraise_kj_exception
        cbool isParseError() except +reraise_kj_exception
        void getIllegalArgument() except +reraise_kj_exception
        cbool isIllegalArgument() except +reraise_kj_exception
    cpdef enum class Exception_Which "Exception::Which"(uint16_t):
        NO_SUCH_TACTIC,
        MISMATCHED_ARGUMENTS,
        PARSE_ERROR,
        ILLEGAL_ARGUMENT,

    Schema getExecutionResult_Schema"capnp::Schema::from<ExecutionResult>"()
    cdef cppclass C_ExecutionResult_Reader "ExecutionResult::Reader":
        ExecutionResult_Which which() except +reraise_kj_exception
        void getFailure() except +reraise_kj_exception
        cbool isFailure() except +reraise_kj_exception
        void getComplete() except +reraise_kj_exception
        cbool isComplete() except +reraise_kj_exception
        C_ExecutionResult_NewState_Reader getNewState() except +reraise_kj_exception
        cbool isNewState() except +reraise_kj_exception
        C_Exception_Reader getProtocolError() except +reraise_kj_exception
        cbool isProtocolError() except +reraise_kj_exception
        cbool hasProtocolError() except +reraise_kj_exception
    cpdef enum class ExecutionResult_Which "ExecutionResult::Which"(uint16_t):
        FAILURE,
        COMPLETE,
        NEW_STATE,
        PROTOCOL_ERROR,

    Schema getExecutionResult_NewState_Schema"capnp::Schema::from<ExecutionResult::NewState>"()
    cdef cppclass C_ExecutionResult_NewState_Reader "ExecutionResult::NewState::Reader":
        C_Graph_Reader getGraph() except +reraise_kj_exception
        cbool hasGraph() except +reraise_kj_exception
        C_ProofState_Reader getState() except +reraise_kj_exception
        cbool hasState() except +reraise_kj_exception

    Schema getPredictionProtocol_Schema"capnp::Schema::from<PredictionProtocol>"()
    cdef cppclass C_PredictionProtocol_Reader "PredictionProtocol::Reader":
        pass

    Schema getPredictionProtocol_Request_Schema"capnp::Schema::from<PredictionProtocol::Request>"()
    cdef cppclass C_PredictionProtocol_Request_Reader "PredictionProtocol::Request::Reader":
        PredictionProtocol_Request_Which which() except +reraise_kj_exception
        C_PredictionProtocol_Request_Initialize_Reader getInitialize() except +reraise_kj_exception
        cbool isInitialize() except +reraise_kj_exception
        C_PredictionProtocol_Request_Predict_Reader getPredict() except +reraise_kj_exception
        cbool isPredict() except +reraise_kj_exception
        uint64_t getSynchronize() except +reraise_kj_exception
        cbool isSynchronize() except +reraise_kj_exception
        void getCheckAlignment() except +reraise_kj_exception
        cbool isCheckAlignment() except +reraise_kj_exception
    cpdef enum class PredictionProtocol_Request_Which "PredictionProtocol::Request::Which"(uint16_t):
        INITIALIZE,
        PREDICT,
        SYNCHRONIZE,
        CHECK_ALIGNMENT,

    Schema getPredictionProtocol_Request_Initialize_Schema"capnp::Schema::from<PredictionProtocol::Request::Initialize>"()
    cdef cppclass C_PredictionProtocol_Request_Initialize_Reader "PredictionProtocol::Request::Initialize::Reader":
        C_DataVersion_Reader getDataVersion() except +reraise_kj_exception
        cbool hasDataVersion() except +reraise_kj_exception
        int32_t getStackSize() except +reraise_kj_exception
        C_AbstractTactic_Reader_List getTactics() except +reraise_kj_exception
        cbool hasTactics() except +reraise_kj_exception
        C_Graph_Reader getGraph() except +reraise_kj_exception
        cbool hasGraph() except +reraise_kj_exception
        StringPtr getLogAnnotation() except +reraise_kj_exception
        cbool hasLogAnnotation() except +reraise_kj_exception
        uint32_t getRepresentative() except +reraise_kj_exception

    Schema getPredictionProtocol_Request_Predict_Schema"capnp::Schema::from<PredictionProtocol::Request::Predict>"()
    cdef cppclass C_PredictionProtocol_Request_Predict_Reader "PredictionProtocol::Request::Predict::Reader":
        C_Graph_Reader getGraph() except +reraise_kj_exception
        cbool hasGraph() except +reraise_kj_exception
        C_ProofState_Reader getState() except +reraise_kj_exception
        cbool hasState() except +reraise_kj_exception

    Schema getPredictionProtocol_Prediction_Schema"capnp::Schema::from<PredictionProtocol::Prediction>"()
    cdef cppclass C_PredictionProtocol_Prediction_Reader "PredictionProtocol::Prediction::Reader":
        C_Tactic_Reader getTactic() except +reraise_kj_exception
        cbool hasTactic() except +reraise_kj_exception
        C_Argument_Reader_List getArguments() except +reraise_kj_exception
        cbool hasArguments() except +reraise_kj_exception
        double getConfidence() except +reraise_kj_exception

    Schema getPredictionProtocol_TextPrediction_Schema"capnp::Schema::from<PredictionProtocol::TextPrediction>"()
    cdef cppclass C_PredictionProtocol_TextPrediction_Reader "PredictionProtocol::TextPrediction::Reader":
        StringPtr getTacticText() except +reraise_kj_exception
        cbool hasTacticText() except +reraise_kj_exception
        double getConfidence() except +reraise_kj_exception

    Schema getPredictionProtocol_Response_Schema"capnp::Schema::from<PredictionProtocol::Response>"()
    cdef cppclass C_PredictionProtocol_Response_Reader "PredictionProtocol::Response::Reader":
        PredictionProtocol_Response_Which which() except +reraise_kj_exception
        void getInitialized() except +reraise_kj_exception
        cbool isInitialized() except +reraise_kj_exception
        C_PredictionProtocol_Prediction_Reader_List getPrediction() except +reraise_kj_exception
        cbool isPrediction() except +reraise_kj_exception
        cbool hasPrediction() except +reraise_kj_exception
        C_PredictionProtocol_TextPrediction_Reader_List getTextPrediction() except +reraise_kj_exception
        cbool isTextPrediction() except +reraise_kj_exception
        cbool hasTextPrediction() except +reraise_kj_exception
        uint64_t getSynchronized() except +reraise_kj_exception
        cbool isSynchronized() except +reraise_kj_exception
        C_PredictionProtocol_Response_Alignment_Reader getAlignment() except +reraise_kj_exception
        cbool isAlignment() except +reraise_kj_exception
    cpdef enum class PredictionProtocol_Response_Which "PredictionProtocol::Response::Which"(uint16_t):
        INITIALIZED,
        PREDICTION,
        TEXT_PREDICTION,
        SYNCHRONIZED,
        ALIGNMENT,

    Schema getPredictionProtocol_Response_Alignment_Schema"capnp::Schema::from<PredictionProtocol::Response::Alignment>"()
    cdef cppclass C_PredictionProtocol_Response_Alignment_Reader "PredictionProtocol::Response::Alignment::Reader":
        C_Int64_List getUnalignedTactics() except +reraise_kj_exception
        cbool hasUnalignedTactics() except +reraise_kj_exception
        C_Node_Reader_List getUnalignedDefinitions() except +reraise_kj_exception
        cbool hasUnalignedDefinitions() except +reraise_kj_exception

    Schema getConflatableEdges_Schema"capnp::Schema::from<ConflatableEdges>"()
    cdef cppclass C_ConflatableEdges_Reader "ConflatableEdges::Reader":
        pass
    cpdef enum class EdgeClassification "capnp::schemas::EdgeClassification_e40ecefe7a595f8e"(uint16_t):
        CONTEXT_ELEM,
        CONTEXT_SUBJECT,
        CONTEXT_DEF_TYPE,
        CONTEXT_DEF_TERM,
        CONST_TYPE,
        CONST_UNDEF,
        CONST_DEF,
        CONST_OPAQUE_DEF,
        CONST_PRIMITIVE,
        IND_TYPE,
        IND_CONSTRUCT,
        IND_PROJECTION,
        PROJ_TERM,
        CONSTRUCT_TERM,
        CAST_TERM,
        CAST_TYPE,
        PROD_TYPE,
        PROD_TERM,
        LAMBDA_TYPE,
        LAMBDA_TERM,
        LET_IN_DEF,
        LET_IN_TYPE,
        LET_IN_TERM,
        APP_FUN,
        APP_ARG,
        CASE_TERM,
        CASE_RETURN,
        CASE_BRANCH_POINTER,
        CASE_IND,
        C_B_CONSTRUCT,
        C_B_TERM,
        FIX_MUTUAL,
        FIX_RETURN,
        FIX_FUN_TYPE,
        FIX_FUN_TERM,
        CO_FIX_MUTUAL,
        CO_FIX_RETURN,
        CO_FIX_FUN_TYPE,
        CO_FIX_FUN_TERM,
        REL_POINTER,
        EVAR_SUBST_POINTER,
        EVAR_SUBST_TERM,
        EVAR_SUBST_TARGET,
        EVAR_SUBJECT,

    cdef cppclass C_DynamicStruct_Reader" ::capnp::DynamicStruct::Reader":
        C_Graph_Node_Reader asGraph_Node"as<Graph::Node>"()
        C_Graph_EdgeTarget_Reader asGraph_EdgeTarget"as<Graph::EdgeTarget>"()
        C_Graph_Reader asGraph"as<Graph>"()
        C_IntP_Reader asIntP"as<IntP>"()
        C_Definition_Reader asDefinition"as<Definition>"()
        C_FloatP_Reader asFloatP"as<FloatP>"()
        C_Graph_Node_Label_Reader asGraph_Node_Label"as<Graph::Node::Label>"()
        C_Definition_Status_Substituted_Reader asDefinition_Status_Substituted"as<Definition::Status::Substituted>"()
        C_Definition_Status_Reader asDefinition_Status"as<Definition::Status>"()
        C_ProofStep_Reader asProofStep"as<ProofStep>"()
        C_Tactic_Reader asTactic"as<Tactic>"()
        C_ProofStep_Tactic_Reader asProofStep_Tactic"as<ProofStep::Tactic>"()
        C_Outcome_Reader asOutcome"as<Outcome>"()
        C_Outcome_Term_Reader asOutcome_Term"as<Outcome::Term>"()
        C_ProofState_Reader asProofState"as<ProofState>"()
        C_Argument_Reader asArgument"as<Argument>"()
        C_ProofState_Root_Reader asProofState_Root"as<ProofState::Root>"()
        C_Node_Reader asNode"as<Node>"()
        C_Argument_Term_Reader asArgument_Term"as<Argument::Term>"()
        C_Graph_EdgeTarget_Target_Reader asGraph_EdgeTarget_Target"as<Graph::EdgeTarget::Target>"()
        C_AbstractTactic_Reader asAbstractTactic"as<AbstractTactic>"()
        C_DataVersion_Reader asDataVersion"as<DataVersion>"()
        C_Dataset_Reader asDataset"as<Dataset>"()
        C_Exception_Reader asException"as<Exception>"()
        C_ExecutionResult_Reader asExecutionResult"as<ExecutionResult>"()
        C_ExecutionResult_NewState_Reader asExecutionResult_NewState"as<ExecutionResult::NewState>"()
        C_PredictionProtocol_Reader asPredictionProtocol"as<PredictionProtocol>"()
        C_PredictionProtocol_Request_Reader asPredictionProtocol_Request"as<PredictionProtocol::Request>"()
        C_PredictionProtocol_Request_Initialize_Reader asPredictionProtocol_Request_Initialize"as<PredictionProtocol::Request::Initialize>"()
        C_PredictionProtocol_Request_Predict_Reader asPredictionProtocol_Request_Predict"as<PredictionProtocol::Request::Predict>"()
        C_PredictionProtocol_Prediction_Reader asPredictionProtocol_Prediction"as<PredictionProtocol::Prediction>"()
        C_PredictionProtocol_TextPrediction_Reader asPredictionProtocol_TextPrediction"as<PredictionProtocol::TextPrediction>"()
        C_PredictionProtocol_Response_Reader asPredictionProtocol_Response"as<PredictionProtocol::Response>"()
        C_PredictionProtocol_Response_Alignment_Reader asPredictionProtocol_Response_Alignment"as<PredictionProtocol::Response::Alignment>"()
        C_ConflatableEdges_Reader asConflatableEdges"as<ConflatableEdges>"()

cdef extern from "capnp/list.h":
    cdef cppclass C_Uint16_List " ::capnp::List<uint16_t, ::capnp::Kind::PRIMITIVE>::Reader":
        uint16_t operator[](uint) except +reraise_kj_exception
        uint size()
    cdef cppclass C_Uint32_List " ::capnp::List<uint32_t, ::capnp::Kind::PRIMITIVE>::Reader":
        uint32_t operator[](uint) except +reraise_kj_exception
        uint size()
    cdef cppclass C_Uint64_List " ::capnp::List<uint64_t, ::capnp::Kind::PRIMITIVE>::Reader":
        uint64_t operator[](uint) except +reraise_kj_exception
        uint size()
    cdef cppclass C_Int64_List " ::capnp::List<int64_t, ::capnp::Kind::PRIMITIVE>::Reader":
        int64_t operator[](uint) except +reraise_kj_exception
        uint size()
    cdef cppclass C_String_List " ::capnp::List<capnp::Text, ::capnp::Kind::BLOB>::Reader":
        StringPtr operator[](uint) except +reraise_kj_exception
        uint size()
    cdef cppclass C_PredictionProtocol_Prediction_Reader_List " ::capnp::List<PredictionProtocol::Prediction, ::capnp::Kind::STRUCT>::Reader":
        C_PredictionProtocol_Prediction_Reader operator[](uint) except +reraise_kj_exception
        uint size()
    cdef cppclass C_Graph_Node_Reader_List " ::capnp::List<Graph::Node, ::capnp::Kind::STRUCT>::Reader":
        C_Graph_Node_Reader operator[](uint) except +reraise_kj_exception
        uint size()
    cdef cppclass C_AbstractTactic_Reader_List " ::capnp::List<AbstractTactic, ::capnp::Kind::STRUCT>::Reader":
        C_AbstractTactic_Reader operator[](uint) except +reraise_kj_exception
        uint size()
    cdef cppclass C_ProofStep_Reader_List " ::capnp::List<ProofStep, ::capnp::Kind::STRUCT>::Reader":
        C_ProofStep_Reader operator[](uint) except +reraise_kj_exception
        uint size()
    cdef cppclass C_PredictionProtocol_TextPrediction_Reader_List " ::capnp::List<PredictionProtocol::TextPrediction, ::capnp::Kind::STRUCT>::Reader":
        C_PredictionProtocol_TextPrediction_Reader operator[](uint) except +reraise_kj_exception
        uint size()
    cdef cppclass C_ProofState_Reader_List " ::capnp::List<ProofState, ::capnp::Kind::STRUCT>::Reader":
        C_ProofState_Reader operator[](uint) except +reraise_kj_exception
        uint size()
    cdef cppclass C_Graph_EdgeTarget_Reader_List " ::capnp::List<Graph::EdgeTarget, ::capnp::Kind::STRUCT>::Reader":
        C_Graph_EdgeTarget_Reader operator[](uint) except +reraise_kj_exception
        uint size()
    cdef cppclass C_Argument_Reader_List " ::capnp::List<Argument, ::capnp::Kind::STRUCT>::Reader":
        C_Argument_Reader operator[](uint) except +reraise_kj_exception
        uint size()
    cdef cppclass C_Outcome_Reader_List " ::capnp::List<Outcome, ::capnp::Kind::STRUCT>::Reader":
        C_Outcome_Reader operator[](uint) except +reraise_kj_exception
        uint size()
    cdef cppclass C_Node_Reader_List " ::capnp::List<Node, ::capnp::Kind::STRUCT>::Reader":
        C_Node_Reader operator[](uint) except +reraise_kj_exception
        uint size()

cdef class Graph_Node_Reader:
    cdef C_Graph_Node_Reader source
    cdef object root
    @staticmethod
    cdef init(C_Graph_Node_Reader source, object root)

cdef class Graph_EdgeTarget_Reader:
    cdef C_Graph_EdgeTarget_Reader source
    cdef object root
    @staticmethod
    cdef init(C_Graph_EdgeTarget_Reader source, object root)

cdef class Graph_Reader:
    cdef C_Graph_Reader source
    cdef object root
    @staticmethod
    cdef init(C_Graph_Reader source, object root)

cdef class IntP_Reader:
    cdef C_IntP_Reader source
    cdef object root
    @staticmethod
    cdef init(C_IntP_Reader source, object root)

cdef class Definition_Reader:
    cdef C_Definition_Reader source
    cdef object root
    @staticmethod
    cdef init(C_Definition_Reader source, object root)

cdef class FloatP_Reader:
    cdef C_FloatP_Reader source
    cdef object root
    @staticmethod
    cdef init(C_FloatP_Reader source, object root)

cdef class Graph_Node_Label_Reader:
    cdef C_Graph_Node_Label_Reader source
    cdef object root
    @staticmethod
    cdef init(C_Graph_Node_Label_Reader source, object root)

cdef class Definition_Status_Substituted_Reader:
    cdef C_Definition_Status_Substituted_Reader source
    cdef object root
    @staticmethod
    cdef init(C_Definition_Status_Substituted_Reader source, object root)

cdef class Definition_Status_Reader:
    cdef C_Definition_Status_Reader source
    cdef object root
    @staticmethod
    cdef init(C_Definition_Status_Reader source, object root)

cdef class ProofStep_Reader:
    cdef C_ProofStep_Reader source
    cdef object root
    @staticmethod
    cdef init(C_ProofStep_Reader source, object root)

cdef class Tactic_Reader:
    cdef C_Tactic_Reader source
    cdef object root
    @staticmethod
    cdef init(C_Tactic_Reader source, object root)

cdef class ProofStep_Tactic_Reader:
    cdef C_ProofStep_Tactic_Reader source
    cdef object root
    @staticmethod
    cdef init(C_ProofStep_Tactic_Reader source, object root)

cdef class Outcome_Reader:
    cdef C_Outcome_Reader source
    cdef object root
    @staticmethod
    cdef init(C_Outcome_Reader source, object root)

cdef class Outcome_Term_Reader:
    cdef C_Outcome_Term_Reader source
    cdef object root
    @staticmethod
    cdef init(C_Outcome_Term_Reader source, object root)

cdef class ProofState_Reader:
    cdef C_ProofState_Reader source
    cdef object root
    @staticmethod
    cdef init(C_ProofState_Reader source, object root)

cdef class Argument_Reader:
    cdef C_Argument_Reader source
    cdef object root
    @staticmethod
    cdef init(C_Argument_Reader source, object root)

cdef class ProofState_Root_Reader:
    cdef C_ProofState_Root_Reader source
    cdef object root
    @staticmethod
    cdef init(C_ProofState_Root_Reader source, object root)

cdef class Node_Reader:
    cdef C_Node_Reader source
    cdef object root
    @staticmethod
    cdef init(C_Node_Reader source, object root)

cdef class Argument_Term_Reader:
    cdef C_Argument_Term_Reader source
    cdef object root
    @staticmethod
    cdef init(C_Argument_Term_Reader source, object root)

cdef class Graph_EdgeTarget_Target_Reader:
    cdef C_Graph_EdgeTarget_Target_Reader source
    cdef object root
    @staticmethod
    cdef init(C_Graph_EdgeTarget_Target_Reader source, object root)

cdef class AbstractTactic_Reader:
    cdef C_AbstractTactic_Reader source
    cdef object root
    @staticmethod
    cdef init(C_AbstractTactic_Reader source, object root)

cdef class DataVersion_Reader:
    cdef C_DataVersion_Reader source
    cdef object root
    @staticmethod
    cdef init(C_DataVersion_Reader source, object root)

cdef class Dataset_Reader:
    cdef C_Dataset_Reader source
    cdef object root
    @staticmethod
    cdef init(C_Dataset_Reader source, object root)

cdef class Exception_Reader:
    cdef C_Exception_Reader source
    cdef object root
    @staticmethod
    cdef init(C_Exception_Reader source, object root)

cdef class ExecutionResult_Reader:
    cdef C_ExecutionResult_Reader source
    cdef object root
    @staticmethod
    cdef init(C_ExecutionResult_Reader source, object root)

cdef class ExecutionResult_NewState_Reader:
    cdef C_ExecutionResult_NewState_Reader source
    cdef object root
    @staticmethod
    cdef init(C_ExecutionResult_NewState_Reader source, object root)

cdef class PredictionProtocol_Reader:
    cdef C_PredictionProtocol_Reader source
    cdef object root
    @staticmethod
    cdef init(C_PredictionProtocol_Reader source, object root)

cdef class PredictionProtocol_Request_Reader:
    cdef C_PredictionProtocol_Request_Reader source
    cdef object root
    @staticmethod
    cdef init(C_PredictionProtocol_Request_Reader source, object root)

cdef class PredictionProtocol_Request_Initialize_Reader:
    cdef C_PredictionProtocol_Request_Initialize_Reader source
    cdef object root
    @staticmethod
    cdef init(C_PredictionProtocol_Request_Initialize_Reader source, object root)

cdef class PredictionProtocol_Request_Predict_Reader:
    cdef C_PredictionProtocol_Request_Predict_Reader source
    cdef object root
    @staticmethod
    cdef init(C_PredictionProtocol_Request_Predict_Reader source, object root)

cdef class PredictionProtocol_Prediction_Reader:
    cdef C_PredictionProtocol_Prediction_Reader source
    cdef object root
    @staticmethod
    cdef init(C_PredictionProtocol_Prediction_Reader source, object root)

cdef class PredictionProtocol_TextPrediction_Reader:
    cdef C_PredictionProtocol_TextPrediction_Reader source
    cdef object root
    @staticmethod
    cdef init(C_PredictionProtocol_TextPrediction_Reader source, object root)

cdef class PredictionProtocol_Response_Reader:
    cdef C_PredictionProtocol_Response_Reader source
    cdef object root
    @staticmethod
    cdef init(C_PredictionProtocol_Response_Reader source, object root)

cdef class PredictionProtocol_Response_Alignment_Reader:
    cdef C_PredictionProtocol_Response_Alignment_Reader source
    cdef object root
    @staticmethod
    cdef init(C_PredictionProtocol_Response_Alignment_Reader source, object root)

cdef class ConflatableEdges_Reader:
    cdef C_ConflatableEdges_Reader source
    cdef object root
    @staticmethod
    cdef init(C_ConflatableEdges_Reader source, object root)

cdef class Uint16_List:
    cdef C_Uint16_List source
    cdef object root
    @staticmethod
    cdef init(C_Uint16_List source, object root)

cdef class Uint32_List:
    cdef C_Uint32_List source
    cdef object root
    @staticmethod
    cdef init(C_Uint32_List source, object root)

cdef class Uint64_List:
    cdef C_Uint64_List source
    cdef object root
    @staticmethod
    cdef init(C_Uint64_List source, object root)

cdef class Int64_List:
    cdef C_Int64_List source
    cdef object root
    @staticmethod
    cdef init(C_Int64_List source, object root)

cdef class String_List:
    cdef C_String_List source
    cdef object root
    @staticmethod
    cdef init(C_String_List source, object root)
cdef class PredictionProtocol_Prediction_Reader_List:
    cdef C_PredictionProtocol_Prediction_Reader_List source
    cdef object root
    @staticmethod
    cdef init(C_PredictionProtocol_Prediction_Reader_List source, object root)
cdef class Graph_Node_Reader_List:
    cdef C_Graph_Node_Reader_List source
    cdef object root
    @staticmethod
    cdef init(C_Graph_Node_Reader_List source, object root)
cdef class AbstractTactic_Reader_List:
    cdef C_AbstractTactic_Reader_List source
    cdef object root
    @staticmethod
    cdef init(C_AbstractTactic_Reader_List source, object root)
cdef class ProofStep_Reader_List:
    cdef C_ProofStep_Reader_List source
    cdef object root
    @staticmethod
    cdef init(C_ProofStep_Reader_List source, object root)
cdef class PredictionProtocol_TextPrediction_Reader_List:
    cdef C_PredictionProtocol_TextPrediction_Reader_List source
    cdef object root
    @staticmethod
    cdef init(C_PredictionProtocol_TextPrediction_Reader_List source, object root)
cdef class ProofState_Reader_List:
    cdef C_ProofState_Reader_List source
    cdef object root
    @staticmethod
    cdef init(C_ProofState_Reader_List source, object root)
cdef class Graph_EdgeTarget_Reader_List:
    cdef C_Graph_EdgeTarget_Reader_List source
    cdef object root
    @staticmethod
    cdef init(C_Graph_EdgeTarget_Reader_List source, object root)
cdef class Argument_Reader_List:
    cdef C_Argument_Reader_List source
    cdef object root
    @staticmethod
    cdef init(C_Argument_Reader_List source, object root)
cdef class Outcome_Reader_List:
    cdef C_Outcome_Reader_List source
    cdef object root
    @staticmethod
    cdef init(C_Outcome_Reader_List source, object root)
cdef class Node_Reader_List:
    cdef C_Node_Reader_List source
    cdef object root
    @staticmethod
    cdef init(C_Node_Reader_List source, object root)


from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
import sys
import socket
import socketserver
import argparse
import contextlib
from pytact.data_reader import (data_reader, Original, capnp_message_generator, ProofState,
                                TacticPredictionGraph, TacticPredictionsGraph,
                                TacticPredictionText, TacticPredictionsText,
                                GlobalContextMessage, CheckAlignmentMessage, CheckAlignmentResponse)

@dataclass(eq=True, frozen=True)
class GlobalArgument:
    identity: int
@dataclass(eq=True, frozen=True)
class LocalArgument:
    context_index: int
@dataclass(eq=True, frozen=True)
class OracleTactic:
    tactic_id: int
    arguments: tuple[GlobalArgument | LocalArgument, ...]
    clean: bool

def text_prediction_loop(text_oracle_data, context: GlobalContextMessage):
    prediction_requests = context.prediction_requests
    for msg in prediction_requests:
        if isinstance(msg, ProofState):
            proof_state = msg
            if proof_state.text in text_oracle_data:
                preds = [TacticPredictionText(t, 1) for t in text_oracle_data[proof_state.text]]
            else:
                preds = []
            prediction_requests.send(TacticPredictionsText(preds))
        elif isinstance(msg, CheckAlignmentMessage):
            alignment = CheckAlignmentResponse([], [])
            prediction_requests.send(alignment)
        elif isinstance(msg, GlobalContextMessage):
            text_prediction_loop(text_oracle_data, msg)
        else:
            raise Exception("Capnp protocol error")

def graph_prediction_loop(context: GlobalContextMessage, oracle_data, known_definitions, known_tactics):
    available_tacticids = set([ t.ident for t in context.tactics ])
    available_definitions = { d.node.identity : d.node for d in context.definitions.definitions() }
    prediction_requests = context.prediction_requests
    for msg in prediction_requests:
        if isinstance(msg, ProofState):
            proof_state = msg
            def resolve_arg(arg):
                if isinstance(arg, LocalArgument):
                    return proof_state.context[arg.context_index]
                elif isinstance(arg, GlobalArgument) and arg.identity in available_definitions:
                    return available_definitions[arg.identity]
                else:
                    return None
            possible_tactics = [
                TacticPredictionGraph(t.tactic_id,
                                    [resolve_arg(arg) for arg in t.arguments],
                                    1 if t.clean else 0.95)
                for t in sorted(oracle_data[proof_state.root.identity], key = lambda t: not t.clean)
                if t.tactic_id in available_tacticids and
                all([resolve_arg(arg) is not None for arg in t.arguments])]
            prediction_requests.send(TacticPredictionsGraph(possible_tactics))
        elif isinstance(msg, CheckAlignmentMessage):
            unknown_definitions = [ d for d in context.definitions.definitions()
                                    if d.node.identity not in known_definitions ]
            unknown_tactics = [ t.ident for t in context.tactics
                                if t.ident not in known_tactics ]
            alignment = CheckAlignmentResponse(unknown_definitions, unknown_tactics)
            prediction_requests.send(alignment)
        elif isinstance(msg, GlobalContextMessage):
            graph_prediction_loop(msg, oracle_data, known_definitions, known_tactics)
        else:
            raise Exception("Capnp protocol error")

def run_session(oracle_data, text_oracle_data, known_definitions, known_tactics, args, capnp_socket, record_file):
    messages_generator = capnp_message_generator(capnp_socket, record_file)
    if args.mode == 'text':
        print('Python server running in text mode')
        text_prediction_loop(text_oracle_data, messages_generator)
    elif args.mode == 'graph':
        print('Python server running in graph mode')
        graph_prediction_loop(messages_generator, oracle_data, known_definitions, known_tactics)
    else:
        raise Exception("The 'mode' argument needs to be either 'text' or 'graph'")

def main():
    parser = argparse.ArgumentParser(
        description = 'A tactic prediction server acting as an oracle, retrieving it\'s information from a dataset',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('mode',
                        type=str,
                        choices=['graph', 'text'],
                        help='"graph" to communicate in graph-mode, "text" to communicate in text-mode')
    parser.add_argument('dataset',
                        type=str,
                        help=('The location of the dataset from which to extract the oracle information. ' +
                              'Either a dataset directory, or a SquashFS image, ' +
                              'which will be automatically mounted.'))
    parser.add_argument('--tcp',
                        dest='port',
                        type = int,
                        default = None,
                        help='Run in tcp mode instead of stdin mode on the specified port.')
    parser.add_argument('--record',
                        dest="record_file",
                        type = str,
                        default = None,
                        help='Record all exchanged messages to the specified file, so that they can later be ' +
                        'replayed through "pytact-fake-coq"')
    cmd_args = parser.parse_args()

    print("Building oracle data...")
    dataset_path = Path(cmd_args.dataset).resolve()
    oracle_data = defaultdict(set)
    text_oracle_data = defaultdict(set)
    known_definitions = set()
    known_tactics = set()
    with data_reader(dataset_path) as data:
        for datafile in data.values():
            for d in datafile.definitions():
                known_definitions.add(d.node.identity)
                if proof := d.proof:
                    for step in proof:
                        for outcome in step.outcomes:
                            if outcome.tactic is None:
                                continue # If the tactic is unknown we are screwed
                            known_tactics.add(outcome.tactic.ident)
                            if not isinstance(d.status, Original):
                                continue # For an oracle, we are not interested in non-original proofs
                            if len(outcome.after) == 1:
                                if outcome.before.id == outcome.after[0].id:
                                    continue # This tactic didn't do anything, we can ignore it
                                if outcome.before.root.identity == outcome.after[0].root.identity:
                                    # This tactic did something, but very minimally, usually just an identity cast
                                    continue
                            text_oracle_data[outcome.before.text].add(outcome.tactic.text_non_anonymous)
                            tactic_args = outcome.tactic_arguments
                            if any(arg is None for arg in tactic_args):
                                continue # If an argument is unknown we are screwed
                            args = []
                            for arg in tactic_args:
                                if arg_def := arg.definition:
                                    args.append(GlobalArgument(arg_def.node.identity))
                                else:
                                    args.append(LocalArgument(list(outcome.before.context).index(arg)))
                            oracle_tactic = OracleTactic(outcome.tactic.ident, tuple(args),
                                                         outcome.tactic.text == outcome.tactic.interm_text)
                            oracle_data[outcome.before.root.identity].add(oracle_tactic)
    print("Oracle data built, ready for incoming connections")

    if cmd_args.record_file is not None:
        record_context = open(cmd_args.record_file, 'wb')
    else:
        record_context = contextlib.nullcontext()
    with record_context as record_file:
        if cmd_args.port is not None:
            class Handler(socketserver.BaseRequestHandler):
                def handle(self):
                    run_session(oracle_data, text_oracle_data, known_definitions, known_tactics,
                                cmd_args, self.request, record_file)
            class Server(socketserver.ThreadingTCPServer):
                def __init__(self, *kwargs):
                    self.allow_reuse_address = True
                    self.daemon_threads = True
                    super().__init__(*kwargs)
            addr = ('localhost', cmd_args.port)
            with Server(addr, Handler) as server:
                server.serve_forever()
        else:
            capnp_socket = socket.socket(fileno=sys.stdin.fileno())
            run_session(oracle_data, text_oracle_data, known_definitions, known_tactics,
                        cmd_args, capnp_socket, record_file)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import os
import sys
import keyword
from typing import Any
from jinja2 import Environment, PackageLoader
import inflection
import shutil

import capnp
capnp_executable = shutil.which("capnp")
if not capnp_executable:
    raise ValueError("capnp executable not found")
capnp_dir = os.path.dirname(capnp_executable)
include_dir = [os.path.abspath(os.path.join(capnp_dir, "..", "include")),
               os.path.abspath(os.path.join(capnp_dir, "..", "include", "capnp"))]
capnp.add_import_hook(include_dir)
import schema_capnp

def cython_name(name):
    return '_'.join([inflection.camelize(p) for p in name.split(':')[1].split('.')])

def cpp_name(name):
    return '::'.join([inflection.camelize(p) for p in name.split(':')[1].split('.')])

def pycapnp_name(name):
    return '.'.join([inflection.camelize(p) for p in name.split(':')[1].split('.')])

def mk_return(cpp, cython, sort):
    return {'cpp_return': cpp, 'cython_return': cython, 'sort': sort}

def gen_list_type_info(nodes, list_type):
    list_which = list_type.elementType.which()
    if list_which in ['void', 'bool',
                        'int8', 'int16', 'int32', 'int64',
                        'uint8', 'uint16', 'uint32', 'uint64',
                        'float32', 'float64']:
        return list_which.capitalize(), None
    elif list_which == 'text':
        return 'String', None
    elif list_which == 'data':
        return 'String', None
    elif list_which == 'list':
        return None # TODO
    elif list_which == 'enum':
        return None # TODO
    elif list_which == 'struct':
        return (
            cython_name(nodes[list_type.elementType.struct.typeId].displayName)+'_Reader',
            cpp_name(nodes[list_type.elementType.struct.typeId].displayName)
        )
    elif list_which == 'interface':
        return None
    elif list_which == 'anyPointer':
        return None # TODO
    else: assert False

def gen_type_info(nodes, type, list_accum):
    which = type.which()
    if which == 'void':
        return mk_return('void', None, 'native')
    elif which == 'bool':
        return mk_return('cbool', None, 'native')
    elif which in ['int8', 'int16', 'int32', 'int64',
                    'uint8', 'uint16', 'uint32', 'uint64']:
        return mk_return(which+'_t', None, 'native')
    elif which == 'float32':
        return mk_return('float', None, 'native')
    elif which == 'float64':
        return mk_return('double', None, 'native')
    elif which == 'text':
        return mk_return('StringPtr', None, 'string')
    elif which == 'data':
        return mk_return('StringPtr', None, 'string')
    elif which == 'list':
        sub_type = gen_list_type_info(nodes, type.list)
        if sub_type is None:
            return None
        cython_type, cpp_type = sub_type
        if cpp_type is not None:
            list_accum.add((cython_type, cpp_type))
        return mk_return('C_'+cython_type+'_List', cython_type+'_List', 'struct')
    elif which == 'enum':
        name = nodes[type.enum.typeId].displayName
        return mk_return(cython_name(name), None, 'enum')
    elif which == 'struct':
        name = nodes[type.struct.typeId].displayName
        return mk_return('C_'+cython_name(name)+'_Reader', cython_name(name)+'_Reader', 'struct')
    elif which == 'interface':
        return None # TODO
    elif which == 'anyPointer':
        return None # TODO
    else: assert False

def main():
    env = Environment(loader=PackageLoader('generate_api'))

    code = schema_capnp.CodeGeneratorRequest.read(sys.stdin)
    nodes = {node.id: node for node in code.nodes}
    sourceInfo = {node.id: node for node in code.sourceInfo}
    classes = []
    classes_lists = set()
    enums = []
    for node in code.nodes:
        if node.scopeId == 0:
            continue # These are RPC argument structs, which we skip for now
        if node.which() == 'struct':
            struct = node.struct
            c_info = {'cython_name': cython_name(node.displayName),
                      'cpp_name': cpp_name(node.displayName),
                      'pycapnp_name': pycapnp_name(node.displayName),
                      'fields': [],
                      'union_fields': []}
            for field_index, field in enumerate(struct.fields):
                if field.which() == 'slot':
                    f_info: dict[str, Any]
                    if info := gen_type_info(nodes, field.slot.type, classes_lists):
                        f_info = info
                    else:
                        continue
                elif field.which() == 'group':
                    name = nodes[field.group.typeId].displayName
                    f_info = mk_return('C_'+cython_name(name)+'_Reader', cython_name(name)+'_Reader', 'struct')
                else: assert False
                f_info['cpp_name'] = inflection.camelize(field.name)
                cn = inflection.underscore(field.name)
                if cn in keyword.kwlist:
                    cn += '_'
                f_info['cython_name'] = cn
                if schema_capnp.Field.noDiscriminant != field.discriminantValue:
                    f_info['union'] = True
                    c_info['union_fields'].append(inflection.underscore(field.name).upper())
                else:
                    f_info['union'] = False
                f_info['gen_has'] = (field.which() == 'slot' and field.slot.type.which() in
                                     ['text', 'data', 'list', 'struct', 'interface', 'anyPointer'])
                f_info['doc'] = sourceInfo[node.id].members[field_index].docComment.strip()
                c_info['fields'].append(f_info)
            c_info['is_group'] = struct.isGroup
            c_info['doc'] = sourceInfo[node.id].docComment.strip()
            classes.append(c_info)
        elif node.which() == 'enum':
            e_info = {'cython_name': cython_name(node.displayName),
                      'cpp_name': 'capnp::schemas::'+cpp_name(node.displayName)+'_'+f'{node.id:x}',
                      'pycapnp_name': pycapnp_name(node.displayName),
                      'enumerants': [inflection.underscore(e.name).upper() for e in node.enum.enumerants]}
            enums.append(e_info)

    pyx = env.get_template("capnp_cython.pyx")
    pxd = env.get_template("capnp_cython.pxd")
    classes_lists = [{'cython_name': cython_name, 'cpp_name': cpp_name} for cython_name, cpp_name in classes_lists]
    filename = os.path.join(os.path.dirname(__file__), 'graph_api_capnp_cython')
    with open(filename+'.pyx', "w") as out:
        out.write(pyx.render(classes=classes, classes_lists=classes_lists, enums=enums))
    with open(filename+'.pxd', "w") as out:
        out.write(pxd.render(classes=classes, classes_lists=classes_lists, enums=enums))

if __name__ == '__main__':
    main()



import inflection
import graphviz
from pytact.data_reader import ProofState, Node

# Load the cap'n proto library, and the communication specification in 'graph_api.capnp'
import capnp
import pytact.graph_api_capnp as graph_api_capnp

# TODO: Clean up and unify all the functions below

arrow_heads = [ "dot", "inv", "odot", "invdot", "invodot" ]
edge_arrow_map = {}
for group in graph_api_capnp.groupedEdges:
    count = 0
    for sort in group.conflatable:
        edge_arrow_map[sort] = arrow_heads[count]
        count += 1

edge_arrow_map2 = {e.raw : arrow for (e, arrow) in edge_arrow_map.items()}

def visualize_proof_state(state: ProofState):

    dot = graphviz.Digraph()
    dot.attr('graph', ordering="out")

    seen = set()
    nodes_left = 100

    def recurse(node: Node, depth):
        nonlocal seen
        nonlocal nodes_left

        id = str(node)
        if id in seen:
            return id
        seen.add(id)
        nodes_left -= 1
        if nodes_left < 0:
            id = 'trunc' + str(nodes_left)
            dot.node(id, 'truncated')
            return id

        if d := node.definition:
            label = d.name
        else:
            label = inflection.camelize(str(node.label.which.name.lower()))
        dot.node(id, label=label)
        if node.definition:
            depth -= 1
        if depth >= 0:
            for edge, child in node.children:
                cid = recurse(child, depth)
                dot.edge(id, cid, arrowtail=edge_arrow_map2[edge], dir="both")
        return id
    recurse(state.root, 0)

    dot.render(filename='python_graph', view=False, cleanup=True)

def visualize(graph, state, showLabel = False, graph1 = None,
              filename='python_graph', cleanup=True):
    nodes = graph.nodes
    root = state.root
    context = state.context
    assert all(n < len(nodes) for n in context)

    dot = graphviz.Digraph()
    dot.attr('graph', ordering="out")
    for node, value in enumerate(nodes):
        label = value.label.which()
        if node in context: label = str(node) + ': ['+label+']'
        if node == root:
            label = "Root: "+label
        dot.node(str(node), label)
    for node, value in enumerate(nodes):
        for edge in list(graph.edges)[value.childrenIndex:value.childrenIndex+value.childrenCount]:
            if graph1 != None and edge.target.depIndex == 1:
                target = '1#'+str(edge.target.nodeIndex)
                dot.node(target, "Global:"
                         + str(graph1.nodes[edge.target.nodeIndex].label.definition.name))
            else:
                target = str(edge.target.nodeIndex)
            if showLabel:
                label = str(edge.label)
            else:
                label = ""
            dot.edge(str(node), target, label=label,
                     arrowtail=edge_arrow_map[edge.label], dir="both")

    dot.render(filename=filename, view=False, cleanup=cleanup)

def visualize_defs(graph, defs, showLabel=False, filename='python_grapn', cleanup=True):
    nodes = graph.nodes
    assert all(n < len(nodes) for n in defs)
    dot = graphviz.Digraph()
    dot.attr('graph', ordering="out")
    for node, value in enumerate(nodes):
        label = value.label.which()
        if node in defs: label = str(node) + ': ' + value.label.definition.name
        dot.node(str(node), label)
    for node, value in enumerate(nodes):
        for edge in list(graph.edges)[value.childrenIndex:value.childrenIndex+value.childrenCount]:
            if showLabel:
                label = str(edge.label)
            else:
                label = ""
            dot.edge(str(node), str(edge.target.nodeIndex), label=label,
                     arrowtail=edge_arrow_map[edge.label], dir="both")

    dot.render(filename=filename, view=False, cleanup=True)

def visualize_exception(reason, filename='visualize_graph.pdf', cleanup=None):
    dot = graphviz.Digraph()
    dot.node(str(reason), str(reason))
    dot.render(filename, view=False, cleanup=cleanup)


class Visualizer:
    def __init__(self, filename, count, show_labels, cleanup):
        self.filename = filename
        self.cnt = 0
        self.count = count
        self.show_labels = show_labels
        self.cleanup = cleanup
    def _visualize(self, filename, result):
        if result.which() == 'newState':
            visualize(result.newState.graph, result.newState.state,
                         filename=filename, showLabel=self.show_labels, cleanup=self.cleanup)
        else:
            visualize_exception(result, filename=self.filename, cleanup=self.cleanup)

    def render(self, result):
        filename = self.filename
        if self.count:
            filename += str(self.cnt)
            self.cnt += 1

        self._visualize(filename, result)



import socket
import argparse
import signal
from subprocess import Popen
import capnp
import pytact.graph_api_capnp as graph_api_capnp
from pytact.data_reader import capnp_message_generator_from_file_lowlevel

def run_fake_client(server_socket, messages_generator):
    socket_reader = graph_api_capnp.PredictionProtocol.Response.read_multiple_packed(
        server_socket, traversal_limit_in_words=2**64-1)
    for msg in messages_generator:
        msg.dynamic.as_builder().write_packed(server_socket)
        prev_sig = signal.signal(signal.SIGINT, signal.SIG_DFL)  # SIGINT catching OFF
        response = next(socket_reader, None)
        signal.signal(signal.SIGINT, prev_sig)  # SIGINT catching ON
        messages_generator.send(response)

def main():
    parser = argparse.ArgumentParser(
        description = 'A fake Coq client that connects to a prediction server and feeds it a stream of previously ' +
                      'recorded messages.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('data',
                        type=str,
                        help='A file that contains a previously recorded communication sequence between Coq ' +
                             'and a prediction server.')
    parser.add_argument('--check',
                        action=argparse.BooleanOptionalAction,
                        default = True,
                        help='Wether or not to compare the response messages of the server to recorded messages.')
    connect_group = parser.add_mutually_exclusive_group(required=True)
    connect_group.add_argument('--tcp',
                        dest='tcp_location',
                        type = str,
                        default = None,
                        help='Connect to a prediction server on the specified ip:port.')
    connect_group.add_argument('--stdin',
                        dest='executable',
                        type = str,
                        default = None,
                        help='Start the specified prediction server and connect to it through stdin.')
    cmd_args = parser.parse_args()

    with open(cmd_args.data, 'rb') as message_file:
        messages_generator = capnp_message_generator_from_file_lowlevel(message_file, check=cmd_args.check)
        if cmd_args.tcp_location is not None:
            addr, port = cmd_args.tcp_location.split(':')
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as server_socket:
                server_socket.connect((addr, int(port)))
                run_fake_client(server_socket, messages_generator)
        else:
            our_sock, their_sock = socket.socketpair()
            process = Popen(cmd_args.executable, shell=True, stdin=their_sock)
            their_sock.close()
            run_fake_client(our_sock, messages_generator)
            our_sock.close()
            process.communicate()

if __name__ == '__main__':
    main()



import contextlib
import sys
import socket
import socketserver
import argparse
import pytact.graph_visualize as gv
from pytact.data_reader import (capnp_message_generator, ProofState,
                                TacticPredictionGraph, TacticPredictionsGraph,
                                TacticPredictionText, TacticPredictionsText,
                                GlobalContextMessage, CheckAlignmentMessage, CheckAlignmentResponse)

def text_prediction_loop(context : GlobalContextMessage):
    tactics = [ 'idtac "is it working?"', 'idtac "yes it is working!"', 'auto' ]
    prediction_requests = context.prediction_requests
    for msg in prediction_requests:
        if isinstance(msg, ProofState):
            proof_state = msg
            print(proof_state.text)
            preds = [TacticPredictionText(t, 0.5) for t in tactics]
            prediction_requests.send(TacticPredictionsText(preds))
        elif isinstance(msg, CheckAlignmentMessage):
            alignment = CheckAlignmentResponse([], [])
            prediction_requests.send(alignment)
        elif isinstance(msg, GlobalContextMessage):
            text_prediction_loop(msg)
        else:
            raise Exception("Capnp protocol error")

def graph_initialize_loop(context : GlobalContextMessage, level):
    print(f"level {level}")
    for cluster in context.definitions.clustered_definitions(full = False):
        print('cluster:')
        for d in cluster:
            print(f'    {d.name}')
    for t in context.tactics:
        print(t)
    print(context.log_annotation)
    prediction_requests = context.prediction_requests
    for msg in prediction_requests:
        if isinstance(msg, ProofState):
            proof_state = msg
            gv.visualize_proof_state(proof_state)
            zeroArgs = [t.ident for t in context.tactics if t.parameters == 0]
            preds = [TacticPredictionGraph(t, [], 0.5) for t in zeroArgs]
            if len(proof_state.context) > 0:
                oneArg = [t.ident for t in context.tactics if t.parameters == 1]
                hyp_node = proof_state.context[0]
                preds += [TacticPredictionGraph(t, [hyp_node], 0.5) for t in oneArg]
            for d in context.definitions.definitions():
                if d.name == "Coq.Init.Logic.I":
                    oneArg = [t.ident for t in context.tactics if t.parameters == 1]
                    preds += [TacticPredictionGraph(t, [d.node], 0.5) for t in oneArg]
            prediction_requests.send(TacticPredictionsGraph(preds))
        elif isinstance(msg, CheckAlignmentMessage):
            unknown_definitions = list(context.definitions.definitions())
            unknown_tactics = [t.ident for t in context.tactics]
            alignment = CheckAlignmentResponse(unknown_definitions, unknown_tactics)
            prediction_requests.send(alignment)
        elif isinstance(msg, GlobalContextMessage):
            graph_initialize_loop(msg, level + 1)
        else:
            raise Exception("Capnp protocol error")

def run_session(args, capnp_socket, record_file):
    messages_generator = capnp_message_generator(capnp_socket, record_file)
    if args.mode == 'text':
        print('Python server running in text mode')
        text_prediction_loop(messages_generator)
    elif args.mode == 'graph':
        print('Python server running in graph mode')
        graph_initialize_loop(messages_generator, 0)
    else:
        raise Exception("The 'mode' argument needs to be either 'text' or 'graph'")

def main():
    parser = argparse.ArgumentParser(
        description = "Example python server capable of communicating with Coq through Tactician's 'synth' tactic",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('mode',
                        type=str,
                        choices=['graph', 'text'],
                        help='"graph" to communicate in graph-mode, "text" to communicate in text-mode')
    parser.add_argument('--tcp',
                        type = int,
                        default = 0,
                        help='Run in tcp mode instead of stdin mode on the specified port.')
    parser.add_argument('--record',
                        dest="record_file",
                        type = str,
                        default = None,
                        help='Record all exchanged messages to the specified file, so that they can later be ' +
                        'replayed through "pytact-fake-coq"')
    args = parser.parse_args()

    if args.record_file is not None:
        record_context = open(args.record_file, 'wb')
    else:
        record_context = contextlib.nullcontext()
    with record_context as record_file:
        if args.tcp != 0:
            class Handler(socketserver.BaseRequestHandler):
                def handle(self):
                    run_session(args, self.request, record_file)
            class Server(socketserver.ThreadingTCPServer):
                def __init__(self, *kwargs):
                    self.allow_reuse_address = True
                    self.daemon_threads = True
                    super().__init__(*kwargs)
            addr = ('localhost', args.tcp)
            with Server(addr, Handler) as server:
                server.daemon_threads = True
                server.serve_forever()
        else:
            capnp_socket = socket.socket(fileno=sys.stdin.fileno())
            run_session(args, capnp_socket, record_file)

if __name__ == '__main__':
    main()
